/*
 ESP32 Laser MIDI Harp
 
 Harp has 12 laser strings and is controlled
 via a web interface. To control the harp
 select the "Laser Harp" wifi network and
 in a browser navagate to "192.168.4.1" and
 you should see the controller page.

 Use the "WEMOS D1 MINI ESP32" board type
 for programming in the Arduino 2.X IDE

 Concept, Design and Implementation by: Craig A. Lindley
 Last Update: 12/28/2022
 */

#include "WiFi.h"
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "Index_HTML.h"

// Access Point Parameters
#define AP_SSID "Laser Harp"
#define AP_PASSWD ""

// GPIO hardware connections
// Strings numbered left to right
#define STRING_1_PIN 27
#define STRING_2_PIN 25
#define STRING_3_PIN 32
#define STRING_4_PIN 22
#define STRING_5_PIN 21
#define STRING_6_PIN 17
#define STRING_7_PIN 19
#define STRING_8_PIN 18
#define STRING_9_PIN 26
#define STRING_10_PIN 34
#define STRING_11_PIN 33
#define STRING_12_PIN 35

#define NUMBER_OF_HARP_STRINGS 12

uint8_t harpStrings[NUMBER_OF_HARP_STRINGS]{
  STRING_1_PIN, STRING_2_PIN, STRING_3_PIN, STRING_4_PIN,
  STRING_5_PIN, STRING_6_PIN, STRING_7_PIN, STRING_8_PIN,
  STRING_9_PIN, STRING_10_PIN, STRING_11_PIN, STRING_12_PIN
};

/************************************************************************/
/***                         Driver Instances                         ***/
/************************************************************************/

// Set web server port number to 80
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

#include "MIDI_Intfc.h"

// Create MIDI driver instance called DIN_MIDI
MIDI_Intfc DIN_MIDI;

#include "MiddleTier.h"

// Create middle tier instance called mt
MiddleTier mt;

/************************************************************************/
/***                         Misc Functions                           ***/
/************************************************************************/

void keyChanged(uint16_t value) {
  // Serial.printf("Key index: %d\n", value);
  mt.setKey(value);
}

void octaveShift(uint16_t value) {
  //  Serial.printf("Octave shift: %d\n", value);
  mt.setOctaveShift(value);
}

void scaleChanged(uint16_t value) {
  //  Serial.printf("Scale index: %d\n", value);
  mt.setScale(value);
}

void programCategoryChanged(uint16_t value) {
  //  Serial.printf("Program category index: %d\n", value);
  mt.setProgramCategory(value);
}

void programChanged(uint16_t value) {
  // Serial.printf("Program index: %d\n", value);
  mt.setProgram(value);
}

void midiChannelChanged(uint16_t value) {
  //  Serial.printf("MIDI channel: %d\n", value);
  mt.setMIDIChannel(value);
}

void stringOrderChanged(uint16_t value) {
  //  Serial.printf("String order: %d\n", value);
  mt.setStringOrder(value);
}

void volumeChanged(uint16_t value) {
  //  Serial.printf("Volume: %d\n", value);
  mt.setVolume(value);
}

void reverbChanged(uint16_t value) {
  //  Serial.printf("Reverb: %d\n", value);
  mt.setReverb(value);
}

void modulationChanged(uint16_t value) {
  //  Serial.printf("Modulation: %d\n", value);
  mt.setModulation(value);
}

void chorusChanged(uint16_t value) {
  //  Serial.printf("Chorus: %d\n", value);
  mt.setChorus(value);
}

void portamentoChanged(uint16_t value) {
  //  Serial.printf("Portamento: %d\n", value);
  mt.setPortamento(value);
}

// Handle WS message
void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
  char cmdToken[5];
  char argToken[5];
 
  AwsFrameInfo *info = (AwsFrameInfo *)arg;
  if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
    data[len] = 0;
    memset(cmdToken, 0, sizeof(cmdToken));
    memset(argToken, 0, sizeof(argToken));

    // Parse out cmd and argument data
    strncpy(cmdToken, (char *)data, 2);
    strncpy(argToken, ((char *)data) + 3, 3);

    //  Serial.printf("CMD_TOKEN: %s\n", cmdToken);

    // Convert numeric argument to an int
    uint16_t val = atoi(argToken);

    // Process the cmd message
    // Was it a key change message ?
    if (strcmp(cmdToken, "KY") == 0) {
      // A key change message
      keyChanged(val);
    } else if (strcmp(cmdToken, "OS") == 0) {
      // A octave shift message
      octaveShift(val);
    } else if (strcmp(cmdToken, "SC") == 0) {
      // A scale change message
      scaleChanged(val);
    } else if (strcmp(cmdToken, "PC") == 0) {
      // A program category change message
      programCategoryChanged(val);
    } else if (strcmp(cmdToken, "PG") == 0) {
      // A program change message
      programChanged(val);
    } else if (strcmp(cmdToken, "MC") == 0) {
      // A MIDI channel change message
      midiChannelChanged(val);
    } else if (strcmp(cmdToken, "SO") == 0) {
      // A string order change message
      stringOrderChanged(val);
    } else if (strcmp(cmdToken, "VU") == 0) {
      // A volume change message
      volumeChanged(val);
    } else if (strcmp(cmdToken, "RV") == 0) {
      // A reverb change message
      reverbChanged(val);
    } else if (strcmp(cmdToken, "CM") == 0) {
      // A modulation change message
      modulationChanged(val);
    } else if (strcmp(cmdToken, "CH") == 0) {
      // A chorus change message
      chorusChanged(val);
    } else if (strcmp(cmdToken, "PT") == 0) {
      // A portamento change message
      portamentoChanged(val);
    }
  }
}

// Process onEvent event
void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type,
             void *arg, uint8_t *data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT:
      Serial.printf("WebSocket client #%u connected from %s\n", client->id(), client->remoteIP().toString().c_str());
      break;
    case WS_EVT_DISCONNECT:
      Serial.printf("WebSocket client #%u disconnected\n", client->id());
      break;
    case WS_EVT_DATA:
      handleWebSocketMessage(arg, data, len);
      break;
    case WS_EVT_PONG:
    case WS_EVT_ERROR:
      break;
  }
}

void initWebSocket() {
  ws.onEvent(onEvent);
  server.addHandler(&ws);
}

// In this case no placeholders like %ccc% need to be defined
String processor(const String &var) {
  return String();
}

/************************************************************************/
/***                         Program Setup                            ***/
/************************************************************************/

void setup() {

  // Start debugging serial port
  Serial.begin(115200);
  delay(1000);

  Serial.println("\n\nStarting ...\n");

  // Configure GPIO pins for string inputs
  pinMode(STRING_1_PIN, INPUT);
  pinMode(STRING_2_PIN, INPUT);
  pinMode(STRING_3_PIN, INPUT);
  pinMode(STRING_4_PIN, INPUT);
  pinMode(STRING_5_PIN, INPUT);
  pinMode(STRING_6_PIN, INPUT);
  pinMode(STRING_7_PIN, INPUT);
  pinMode(STRING_8_PIN, INPUT);
  pinMode(STRING_9_PIN, INPUT);
  pinMode(STRING_10_PIN, INPUT);
  pinMode(STRING_11_PIN, INPUT);
  pinMode(STRING_12_PIN, INPUT);

  // Configure network Access Point
  Serial.println("Configuring AP");
  WiFi.softAP(AP_SSID);
  Serial.println("AP connected !");
  Serial.print("Laser Harp IP address: ");
  IPAddress ip = WiFi.softAPIP();
  Serial.println(ip);

  // Start MIDI driver
  DIN_MIDI.begin();

  // Select acoustic grand piano from general MIDI
  DIN_MIDI.sendProgramChange(DEFAULT_MIDI_PROGRAM, DEFAULT_MIDI_CHANNEL);

  // Initialize web socket
  initWebSocket();

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
    request->send_P(200, "text/html", index_html, processor);
  });

  // Start the server
  server.begin();
}

/************************************************************************/
/***                         Program Loop                             ***/
/************************************************************************/

void loop() {
  ws.cleanupClients();

  // Scan the harp strings
  mt.scanHarpStrings();
}
