/*
 Minimal MIDI Interface for the ESP32

 Concept, Design and Implementation by: Craig A. Lindley
 Last Update: 12/28/2022
*/

#ifndef MIDIINTFC_H
#define MIDIINTFC_H

#include <HardwareSerial.h>

// Identifies Serial2 UART
#define SERIAL_ID 2

// GPIO pins for serial interface
#define TX_PIN 16
#define RX_PIN 4

// MIDI baud rate
#define MIDI_BAUD_RATE 31250
#define MIDI_SERIAL_FORMAT SERIAL_8N1

// MIDI command values
#define NOTE_OFF_CMD 0x80
#define NOTE_ON_CMD 0x90
#define CONTINUOUS_CONTROLLER_CMD 0xB0
#define PROGRAM_CHANGE_CMD 0xC0

class MIDI_Intfc {

public:

  MIDI_Intfc() {
    // Instantiate the serial interface
    hs = new HardwareSerial(SERIAL_ID);
  }

  // Configure and bring up serial interface
  void begin() {
    hs->begin(MIDI_BAUD_RATE, MIDI_SERIAL_FORMAT, RX_PIN, TX_PIN);
  }

  // Send a note on message
  void sendNoteOn(uint8_t note, uint8_t velocity, uint8_t channel) {

    // Serial.printf("Note On: %d\n", note);

    hs->write(NOTE_ON_CMD | (channel - 1));
    hs->write(note);
    hs->write(velocity);
  }

  // Send a note off message
  void sendNoteOff(uint8_t note, uint8_t velocity, uint8_t channel) {

    // Serial.printf("Note Off: %d\n", note);

    hs->write(NOTE_OFF_CMD | (channel - 1));
    hs->write(note);
    hs->write(velocity);
  }

  // Send a CC change message
  void sendControlChange(uint8_t ccCode, uint8_t ccValue, uint8_t channel) {

    // Serial.printf("Control Change - code: %d, value: %d, channel: %d\n", ccCode, ccValue, channel);

    hs->write(CONTINUOUS_CONTROLLER_CMD | (channel - 1));
    hs->write(ccCode);
    hs->write(ccValue);
  }

  // Send a program change message
  void sendProgramChange(uint8_t program, uint8_t channel) {

    // Serial.printf("Program Change - program: %d, channel: %d\n", program, channel);

    hs->write(PROGRAM_CHANGE_CMD | (channel - 1));
    hs->write(program);
  }

private:
  HardwareSerial *hs;
};

#endif
