/*
 Middle Tier Functions for Laser Harp

 Concept, Design and Implementation by: Craig A. Lindley
 Last Update: 12/28/2022
*/

#ifndef MIDDLETIER_H
#define MIDDLETIER_H

// Define defaults
#define MIDDLE_C_MIDI_NOTE_NUMBER 60  // C4 or middle C

#define DEFAULT_MIDI_CHANNEL 1          // Default MIDI channel to xmit on
#define DEFAULT_MIDI_VELOCITY 64        // Default velocity of noteOn messages
#define DEFAULT_OCTAVE_SHIFT 0          // 0 shift keyboard at C4
#define DEFAULT_MIDI_PROGRAM 0          // Default program is piano
#define DEFAULT_PROGRAM_CATEGORY 0      // Default program category
#define DEFAULT_KEY_OFFSET 0            // Default key offset from C
#define DEFAULT_SCALE_INDEX 0           // Default scale is chromatic
#define DEFAULT_STRING_ORDER_LTOR true  // Default string order left to right

// Control change messages
#define CC_VOLUME 7              // Volume CC
#define CC_REVERB 91             // Reverb CC
#define CC_MODULATION 1          // Modulation CC
#define CC_CHORUS 93             // chorus CC
#define CC_PORTAMENTO_TIME 5     // Portamento time CC
#define CC_PORTAMENTO_ON_OFF 65  // Portamento on off

// Macros for handling 16 bit bit fields
#define bitRead16(value, bit) (((value) >> (bit)) & 0x01)
#define bitSet16(value, bit) ((value) |= (1 << (bit)))
#define bitClear16(value, bit) ((value) &= ~(1 << (bit)))

class MiddleTier {
public:

  MiddleTier() {
    // Initialize default values
    midiChannel = DEFAULT_MIDI_CHANNEL;
    midiVelocity = DEFAULT_MIDI_VELOCITY;
    octaveShift = DEFAULT_OCTAVE_SHIFT;
    programCategory = DEFAULT_PROGRAM_CATEGORY;
    keyOffsetFromC = DEFAULT_KEY_OFFSET;
    scaleIndex = DEFAULT_SCALE_INDEX;
    leftToRight = DEFAULT_STRING_ORDER_LTOR;
    stringStates = 0;
  }

  // Scan the harp strings
  void scanHarpStrings() {
    uint8_t stringNumber, midiNoteNumber;

    // Scan the 12 harp strings
    for (int s = 0; s < NUMBER_OF_HARP_STRINGS; s++) {
      // Check strings orientation
      if (leftToRight) {
        stringNumber = s;
      } else {
        stringNumber = NUMBER_OF_HARP_STRINGS - 1 - s;
      }

      // Calculate the MIDI note number
      midiNoteNumber =
        MIDDLE_C_MIDI_NOTE_NUMBER + keyOffsetFromC + (octaveShift * 12) + scaleIntervals[scaleIndex][stringNumber];

      // Now read the harp string's state
      uint8_t state = digitalRead(harpStrings[s]);

      // Is this string active (high) ?
      if (state == HIGH) {
        //      if (state == LOW) {  // For testing purposes
        // Was this note previously on ? If so ignore
        if (bitRead16(stringStates, stringNumber) == LOW) {
          // Display the calculated note number
          //          Serial.printf("NoteOn MIDI note number: %d\n", midiNoteNumber);

          // Yes this string is active so mark it as such
          bitSet16(stringStates, stringNumber);

          // Send a MIDI note on message
          DIN_MIDI.sendNoteOn(midiNoteNumber, midiVelocity, midiChannel);
        }
      } else {
        // Was this note previously on ?
        if (bitRead16(stringStates, stringNumber)) {
          // Display the calculated note number
          //          Serial.printf("NoteOff MIDI note number: %d\n", midiNoteNumber);

          // Yes the note was on so mark as off
          bitClear16(stringStates, stringNumber);

          // Send a MIDI note off message
          DIN_MIDI.sendNoteOff(midiNoteNumber, midiVelocity, midiChannel);
        }
      }
      delay(10);
    }
  }

  // MIDI functions follow

  // Set key
  void setKey(uint16_t key) {
    keyOffsetFromC = key;
  }

  // Set octave shift
  void setOctaveShift(uint16_t os) {
    octaveShift = os - 3;
  }

  // Set scale
  void setScale(uint16_t sc) {
    scaleIndex = sc;
  }

  // Set program category
  void setProgramCategory(uint16_t pc) {
    programCategory = pc;

    // Get comma delimited string of programs in specified category
    // to send to UI
    const char *str = programCategoryNames[pc];

    // Send it to client's UI
    ws.textAll(str);

    // Select the first program of the new set
    setProgram(0);
  }

  // Set program
  void setProgram(uint16_t v) {
    // Calculate program value
    v += programCategory * 8;

    // Select MIDI program
    DIN_MIDI.sendProgramChange(v, midiChannel);
  }

  // Set the current MIDI channel 1 .. 16
  void setMIDIChannel(uint16_t mc) {
    // Serial.printf("MIDI channel: %d\n", mc);
    midiChannel = mc;
  }

  // Set the string order L to R or R to L
  void setStringOrder(uint16_t so) {
    leftToRight = (so == 0) ? true : false;
  }

  // Set volume
  void setVolume(uint16_t v) {
    // Serial.printf("Volume Change: %d\n", v);
    DIN_MIDI.sendControlChange(CC_VOLUME, v, midiChannel);
  }

  // Set reverb
  void setReverb(uint16_t v) {
    // Serial.printf("Reverb Change: %d\n", v);
    DIN_MIDI.sendControlChange(CC_REVERB, v, midiChannel);
  }

  // Set modulation
  void setModulation(uint16_t v) {
    // Serial.printf("Modulation Change: %d\n", v);
    DIN_MIDI.sendControlChange(CC_MODULATION, v, midiChannel);
  }

  // Set chorus
  void setChorus(uint16_t v) {
    // Serial.printf("Chorus Change: %d\n", v);
    DIN_MIDI.sendControlChange(CC_CHORUS, v, midiChannel);
  }

  // Set portamento time
  void setPortamento(uint16_t v) {

    // Serial.printf("Portamento Change: %d\n", v);

    if (v == 0) {
      // If value is 0 turn off portamento
      DIN_MIDI.sendControlChange(CC_PORTAMENTO_ON_OFF, 0, midiChannel);
    } else {
      // If value non 0 turn on portamento and then set value
      DIN_MIDI.sendControlChange(CC_PORTAMENTO_ON_OFF, 127, midiChannel);
      DIN_MIDI.sendControlChange(CC_PORTAMENTO_TIME, v, midiChannel);
    }
  }

private:

  // Private class data
  uint8_t midiChannel;
  uint8_t midiVelocity;
  uint16_t stringStates;
  boolean leftToRight;
  uint8_t keyOffsetFromC;
  int8_t octaveShift;
  uint8_t scaleIndex;
  uint8_t programCategory;

  // Programs 1-8: Piano
  const char *pianoNames =
    "Acoustic Grand,Bright Acoustic,Electric Grand,Honky Tonk,Electric 1,Electric 2,Harpsichord, Clavinet";

  // Programs 9-16: Chromatic Percussion
  const char *chromaticPercussionNames =
    "Celesta,Glockenspiel, Music Box,Vibraphone,Marimba,Xylophone,Tubular Bells,Dulcimer";

  // Programs 17-24: Organ
  const char *organNames =
    "Drawbar,Percussive,Rock,Church, Reed,Accordion,Harmonica,Tango Accordion";

  // Programs 25-32: Guitar
  const char *guitarNames =
    "Acoustic (nylon),Acoustic (steel),Electric (jazz),Electric (clean),Electric (muted),Overdriven,Distortion,Guitar Harmonics";

  // Programs 33-40: Bass
  const char *bassNames =
    "Acoustic,Electric (finger),Electric (pick),Fretless,Slap 1,Slap 2,Synth 1,Synth 2";

  // Programs 41-48: Strings
  const char *stringNames =
    "Violin,Viola,Cello,Contrabass,Tremolo,Pizzicato,Orchestral Harp,Timpani";

  // Programs 49-56: Ensemble
  const char *ensembleNames =
    "String 1,String 2,SynthStrings 1,SynthStrings 2,Choir Aahs,program Oohs,Synth program,Orchestra Hit";

  // Programs 57-64: Brass
  const char *brassNames =
    "Trumpet,Trombone,Tuba,Muted Trumpet,French Horn,Brass Section,SynthBrass 1,SynthBrass 2";

  // Programs 65-72: Reed
  const char *reedNames =
    "Soprano Sax,Alto Sax,Tenor Sax,Baritone Sax,Oboe,English Horn,Bassoon,Clarinet";

  // Programs 73-80: Pipe
  const char *pipeNames =
    "Piccolo,Flute,Recorder,Pan Flute,Blown Bottle,Shakuhachi,Whistle,Ocarina";

  // Programs 81-88: Synth Lead
  const char *synthLeadNames =
    "Lead 1 (square),Lead 2 (sawtooth),Lead 3 (calliope),Lead 4 (chiff),Lead 5 (charang),Lead 6 (program),Lead7 (fifths), Lead 8 (bass + lead)";

  // Programs 89-96: Synth Pad
  const char *synthPadNames =
    "Pad 1 (new age),Pad 2 (warm),Pad 3 (polysynth),Pad 4 (choir),Pad 5 (bowed),Pad 6 (metallic),Pad 7 (halo),Pad 8 (sweep)";

  // Programs 97-104: Synth Effects
  const char *synthEffectNames =
    "FX 1 (rain),FX 2 (soundtrack),FX 3 (crystal),FX 4 (atmosphere),FX 5 (brightness),FX 6 (goblins),FX 7 (echoes),FX 8 (sci-fi)";

  // Programs 105-112: Ethnic
  const char *ethnicNames =
    "Sitar,Banjo,Shamisen,Koto,Kalimba,Bag pipe,Fiddle,Shanai";

  // Programs 113-120: Percussive
  const char *percussiveNames =
    "Tinkle Bell,Agogo,Steel Drums,Woodblock,Taiko Drum,Melodic Tom,Synth Drum,Reverse Cymbal";

  // Programs 121-128: Sound Effects
  const char *soundEffectNames =
    "Guitar Fret Noise,Breath Noise,Seashore,Bird Tweet,Telephone Ring,Helicopter,Applause,Gunshot";

  // Array of arrays of program names
  const char *programCategoryNames[16]{
    pianoNames, chromaticPercussionNames, organNames, guitarNames,
    bassNames, stringNames, ensembleNames, brassNames,
    reedNames, pipeNames, synthLeadNames, synthPadNames,
    synthEffectNames, ethnicNames, percussiveNames, soundEffectNames
  };

  // Scale intervals
  uint8_t scaleIntervals[32][12] = {
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 },        // chromatic
    { 0, 2, 4, 5, 7, 9, 11, 12, 14, 16, 17, 19 },    // major/ionian
    { 0, 2, 3, 5, 7, 9, 10, 12, 14, 15, 17, 19 },    // dorian
    { 0, 1, 3, 5, 7, 8, 10, 12, 13, 15, 17, 19 },    // phrygian
    { 0, 2, 4, 6, 7, 9, 11, 12, 14, 16, 18, 19 },    // lydian
    { 0, 2, 4, 5, 7, 9, 10, 12, 14, 16, 17, 19 },    // mixolydian
    { 0, 2, 3, 5, 7, 8, 10, 12, 14, 15, 17, 19 },    // natural minor/aeolian
    { 0, 1, 3, 5, 6, 8, 10, 12, 13, 15, 17, 18 },    // locrian
    { 0, 1, 4, 5, 7, 8, 10, 12, 13, 16, 17, 19 },    // freygish
    { 0, 2, 3, 5, 7, 9, 11, 12, 14, 15, 17, 19 },    // melodic minor
    { 0, 2, 3, 5, 7, 8, 11, 12, 14, 15, 17, 19 },    // harmonic minor
    { 0, 3, 4, 7, 9, 10, 12, 15, 16, 19, 21, 22 },   // major blues
    { 0, 3, 5, 6, 7, 10, 12, 15, 17, 18, 19, 22 },   // minor blues
    { 0, 2, 4, 7, 9, 12, 14, 16, 19, 21, 24, 26 },   // major pentatonic
    { 0, 3, 5, 7, 10, 12, 15, 17, 19, 22, 24, 27 },  // minor pentatonic
    { 0, 2, 3, 5, 6, 8, 9, 11, 12, 14, 15, 17 },     // diminish
    { 0, 1, 3, 4, 6, 7, 9, 10, 12, 13, 15, 16 },     // combi diminish
    { 0, 1, 4, 5, 7, 8, 11, 12, 13, 16, 17, 19 },    // raga bhairav
    { 0, 1, 4, 6, 7, 9, 11, 12, 13, 16, 18, 19 },    // raga gamansrama
    { 0, 1, 3, 6, 7, 8, 11, 12, 13, 15, 18, 19 },    // raga todi
    { 0, 1, 3, 4, 5, 7, 8, 10, 12, 13, 15, 16 },     // spanish
    { 0, 2, 3, 6, 7, 8, 11, 12, 14, 15, 18, 19 },    // gypsy
    { 0, 2, 4, 5, 6, 8, 10, 12, 14, 16, 17, 18 },    // arabian
    { 0, 2, 5, 7, 10, 12, 14, 17, 19, 22, 24, 26 },  // egyptian
    { 0, 2, 3, 7, 9, 12, 14, 15, 19, 21, 24, 26 },   // hawaiian
    { 0, 1, 3, 7, 8, 12, 13, 15, 19, 20, 24, 25 },   // bali pelog
    { 0, 1, 5, 7, 8, 12, 13, 17, 19, 20, 24, 25 },   // japan
    { 0, 4, 5, 7, 11, 12, 16, 17, 19, 23, 24, 28 },  // ryukyu
    { 0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22 },   // wholetone
    { 0, 3, 4, 7, 8, 11, 12, 15, 16, 19, 20, 23 },   // augmented
    { 0, 2, 4, 6, 10, 12, 14, 16, 18, 22, 24, 26 },  // prometheus
    { 0, 1, 4, 6, 7, 10, 12, 13, 16, 18, 19, 22 }    // tritone
  };
};

#endif
